import yaml
from pathlib import Path
from copy import deepcopy


def generate_new_configs_from_key_and_value(
    base_config, key, value, upper_level_key=None
):
    """
    Generate new configs based on a base config and a strategy.

    Args:
        base_config (dict): The base configuration to generate new configs from.
        key (str): The key to change in the base config.
        value (Union[str, list, int]): The value to change the key to.
        upper_level_key (str, optional): The upper level key to change in the base config; useful for dict. Defaults to None.

    Returns:
        list: A list of new configs.
    """
    configs_to_return = []
    if key == "patch_size":
        for new_patch_size in value:
            new_config = deepcopy(base_config)
            new_config["patch_size"] = new_patch_size
            new_config[key] = new_patch_size
            configs_to_return.append(new_config)
    else:
        # extending via lists is easy enough
        if isinstance(value, list):
            for v in value:
                new_config = deepcopy(base_config)
                if upper_level_key is None:
                    new_config[key] = v
                else:
                    new_config[upper_level_key][key] = v
                configs_to_return.append(new_config)
        # dicts require a bit more effort
        elif isinstance(value, dict):
            for k, v in value.items():
                if not isinstance(v, dict):
                    if not configs_to_return:
                        # first round of configs
                        configs_to_return = generate_new_configs_from_key_and_value(
                            base_config, k, v, key
                        )
                    else:
                        # subsequent rounds of configs
                        temp_configs_to_return = []
                        for config in configs_to_return:
                            temp_configs_to_return.extend(
                                generate_new_configs_from_key_and_value(
                                    config, k, v, key
                                )
                            )
                        configs_to_return.extend(temp_configs_to_return)
                else:
                    raise NotImplementedError("Nested dicts are not supported yet.")
    return configs_to_return


def remove_duplicates(configs_list):
    """
    Remove duplicate configs from a list of configs.

    Args:
        configs_list (list): A list of configs.

    Returns:
        list: A list of configs with duplicates removed.
    """
    configs_to_return = []
    for config in configs_list:
        if config not in configs_to_return:
            configs_to_return.append(config)
    return configs_to_return


def config_generator(base_config_path, strategy_path, output_dir):
    """
    Main function that runs the training and inference.

    Args:
        base_config_path (str): Path to baseline configuration.
        strategy_path (str): Strategy to employ when creating new configurations.
        output_dir (str): The output directory.
    """
    base_config = None
    with open(base_config_path, "r") as f:
        base_config = yaml.safe_load(f)

    strategy = None
    with open(strategy_path, "r") as f:
        strategy = yaml.safe_load(f)

    Path(output_dir).mkdir(parents=True, exist_ok=True)

    new_configs_list = []

    for key, value in strategy.items():
        if not new_configs_list:
            # create first round of configs
            new_configs_list = generate_new_configs_from_key_and_value(
                base_config, key, value
            )
            new_configs_list = remove_duplicates(new_configs_list)
        else:
            # create new configs from the previous round of configs
            temp_configs_to_return = []
            for config in new_configs_list:
                temp_configs_to_return.extend(
                    generate_new_configs_from_key_and_value(config, key, value)
                )
            new_configs_list.extend(temp_configs_to_return)
            new_configs_list = remove_duplicates(new_configs_list)

    for i, config in enumerate(new_configs_list):
        with open(f"{output_dir}/config_{i}.yaml", "w") as f:
            yaml.dump(config, f)



import os
import pathlib

import numpy as np
import pandas as pd
import SimpleITK as sitk
import torch
import torchio
from GANDLF.compute.loss_and_metric import get_loss_and_metrics
from GANDLF.compute.step import step
from GANDLF.data.post_process import global_postprocessing_dict
from GANDLF.utils import (
    get_date_time,
    get_filename_extension_sanitized,
    get_unique_timestamp,
    resample_image,
    reverse_one_hot,
    get_ground_truths_and_predictions_tensor,
    print_and_format_metrics,
)
from GANDLF.metrics import overall_stats
from tqdm import tqdm


def validate_network(
    model, valid_dataloader, scheduler, params, epoch=0, mode="validation"
):
    """
    Function to validate a network for a single epoch

    Parameters
    ----------
    model : if parameters["model"]["type"] == torch, this is a torch.model, otherwise this is OV exec_net
        The model to process the input image with, it should support appropriate dimensions.
    valid_dataloader : torch.DataLoader
        The dataloader for the validation epoch
    params : dict
        The parameters passed by the user yaml
    mode: str
        The mode of validation, used to write outputs, if requested

    Returns
    -------
    average_epoch_valid_loss : float
        Validation loss for the current epoch
    average_epoch_valid_metric : dict
        Validation metrics for the current epoch

    """
    print("*" * 20)
    print("Starting " + mode + " : ")
    print("*" * 20)
    # Initialize a few things
    total_epoch_valid_loss = 0
    total_epoch_valid_metric = {}
    average_epoch_valid_metric = {}

    for metric in params["metrics"]:
        if "per_label" in metric:
            total_epoch_valid_metric[metric] = []
        else:
            total_epoch_valid_metric[metric] = 0

    logits_list = []
    subject_id_list = []
    is_classification = params.get("problem_type") == "classification"
    calculate_overall_metrics = (
        (params["problem_type"] == "classification")
        or (params["problem_type"] == "regression")
    ) and mode == "validation"
    is_inference = mode == "inference"

    # automatic mixed precision - https://pytorch.org/docs/stable/amp.html
    if params["verbose"]:
        if params["model"]["amp"]:
            print("Using Automatic mixed precision", flush=True)

    if scheduler is None:
        current_output_dir = params["output_dir"]  # this is in inference mode
    else:  # this is useful for inference
        current_output_dir = os.path.join(params["output_dir"], "output_" + mode)

    if not (is_inference):
        current_output_dir = os.path.join(current_output_dir, str(epoch))

    pathlib.Path(current_output_dir).mkdir(parents=True, exist_ok=True)

    # Set the model to valid
    if params["model"]["type"] == "torch":
        model.eval()

    # # putting stuff in individual arrays for correlation analysis
    # all_targets = []
    # all_predics = []
    if params["medcam_enabled"] and params["model"]["type"] == "torch":
        model.enable_medcam()
        params["medcam_enabled"] = True

    if params["save_output"] or is_inference:
        if params["problem_type"] != "segmentation":
            outputToWrite = "Epoch,SubjectID,PredictedValue\n"
            file_to_write = os.path.join(current_output_dir, "output_predictions.csv")
            if os.path.exists(file_to_write):
                file_to_write = os.path.join(
                    current_output_dir,
                    "output_predictions_" + get_unique_timestamp() + ".csv",
                )

    # get ground truths for classification problem, validation set
    if calculate_overall_metrics:
        (
            ground_truth_array,
            predictions_array,
        ) = get_ground_truths_and_predictions_tensor(params, "validation_data")

    for batch_idx, (subject) in enumerate(
        tqdm(valid_dataloader, desc="Looping over " + mode + " data")
    ):
        if params["verbose"]:
            print("== Current subject:", subject["subject_id"], flush=True)

        # ensure spacing is always present in params and is always subject-specific
        params["subject_spacing"] = None
        if "spacing" in subject:
            params["subject_spacing"] = subject["spacing"]

        # constructing a new dict because torchio.GridSampler requires torchio.Subject,
        # which requires torchio.Image to be present in initial dict, which the loader does not provide
        subject_dict = {}
        label_ground_truth = None
        label_present = False
        # this is when we want the dataloader to pick up properties of GaNDLF's
        # DataLoader, such as pre-processing and augmentations, if appropriate
        if "label" in subject:
            if subject["label"] != ["NA"]:
                subject_dict["label"] = torchio.LabelMap(
                    path=subject["label"]["path"],
                    tensor=subject["label"]["data"].squeeze(0),
                    affine=subject["label"]["affine"].squeeze(0),
                )
                label_present = True
                label_ground_truth = subject_dict["label"]["data"]

        if "value_keys" in params:  # for regression/classification
            for key in params["value_keys"]:
                subject_dict["value_" + key] = subject[key]
                label_ground_truth = torch.cat(
                    [subject[key] for key in params["value_keys"]], dim=0
                )

        for key in params["channel_keys"]:
            subject_dict[key] = torchio.ScalarImage(
                path=subject[key]["path"],
                tensor=subject[key]["data"].squeeze(0),
                affine=subject[key]["affine"].squeeze(0),
            )

        # regression/classification problem AND label is present
        if (params["problem_type"] != "segmentation") and label_present:
            sampler = torchio.data.LabelSampler(params["patch_size"])
            tio_subject = torchio.Subject(subject_dict)
            generator = sampler(tio_subject, num_patches=params["q_samples_per_volume"])
            pred_output = 0
            for patch in generator:
                image = torch.cat(
                    [patch[key][torchio.DATA] for key in params["channel_keys"]], dim=0
                )
                valuesToPredict = torch.cat(
                    [patch["value_" + key] for key in params["value_keys"]], dim=0
                )
                image = image.unsqueeze(0)
                image = image.float().to(params["device"])
                ## special case for 2D
                if image.shape[-1] == 1:
                    image = torch.squeeze(image, -1)
                if params["model"]["type"] == "torch":
                    pred_output += model(image)
                elif params["model"]["type"] == "openvino":
                    pred_output += torch.from_numpy(
                        model(
                            inputs={params["model"]["IO"][0][0]: image.cpu().numpy()}
                        )[params["model"]["IO"][1][0]]
                    )
                else:
                    raise Exception(
                        "Model type not supported. Please only use 'torch' or 'openvino'."
                    )

            pred_output = pred_output.cpu() / params["q_samples_per_volume"]

            if is_inference and is_classification:
                logits_list.append(pred_output)
                subject_id_list.append(subject.get("subject_id")[0])

            if params["save_output"] or is_inference:
                # we divide by scaling factor here because we multiply by it during loss/metric calculation
                outputToWrite += (
                    str(epoch)
                    + ","
                    + subject["subject_id"][0]
                    + ","
                    + str(pred_output.cpu().max().item() / params["scaling_factor"])
                    + "\n"
                )
            final_loss, final_metric = get_loss_and_metrics(
                image, valuesToPredict, pred_output, params
            )

            if calculate_overall_metrics:
                predictions_array[batch_idx] = (
                    torch.argmax(pred_output[0], 0).cpu().item()
                )
            # # Non network validation related
            total_epoch_valid_loss += final_loss.detach().cpu().item()
            for metric in final_metric.keys():
                if isinstance(total_epoch_valid_metric[metric], list):
                    if len(total_epoch_valid_metric[metric]) == 0:
                        total_epoch_valid_metric[metric] = np.array(
                            final_metric[metric]
                        )
                    else:
                        total_epoch_valid_metric[metric] += np.array(
                            final_metric[metric]
                        )
                else:
                    total_epoch_valid_metric[metric] += final_metric[metric]

        else:  # for segmentation problems OR regression/classification when no label is present
            grid_sampler = torchio.inference.GridSampler(
                torchio.Subject(subject_dict),
                params["patch_size"],
                patch_overlap=params["inference_mechanism"]["patch_overlap"],
            )
            patch_loader = torch.utils.data.DataLoader(grid_sampler, batch_size=1)
            aggregator = torchio.inference.GridAggregator(
                grid_sampler,
                overlap_mode=params["inference_mechanism"]["grid_aggregator_overlap"],
            )

            if params["medcam_enabled"]:
                attention_map_aggregator = torchio.inference.GridAggregator(
                    grid_sampler,
                    overlap_mode=params["inference_mechanism"][
                        "grid_aggregator_overlap"
                    ],
                )

            output_prediction = 0  # this is used for regression/classification
            current_patch = 0
            for patches_batch in patch_loader:
                if params["verbose"]:
                    print(
                        "=== Current patch:",
                        current_patch,
                        ", time : ",
                        get_date_time(),
                        ", location :",
                        patches_batch[torchio.LOCATION],
                        flush=True,
                    )
                current_patch += 1
                image = (
                    torch.cat(
                        [
                            patches_batch[key][torchio.DATA]
                            for key in params["channel_keys"]
                        ],
                        dim=1,
                    )
                    .float()
                    .to(params["device"])
                )

                # calculate metrics if ground truth is present
                label = None
                if params["problem_type"] != "segmentation":
                    label = label_ground_truth
                elif "label" in patches_batch:
                    label = patches_batch["label"][torchio.DATA]

                if label is not None:
                    label = label.to(params["device"])
                    if params["verbose"]:
                        print(
                            "=== Validation shapes : label:",
                            label.shape,
                            ", image:",
                            image.shape,
                            flush=True,
                        )

                if is_inference:
                    result = step(model, image, None, params, train=False)
                else:
                    result = step(model, image, label, params, train=True)

                # get the current attention map and add it to its aggregator
                if params["medcam_enabled"]:
                    _, _, output, attention_map = result
                    attention_map_aggregator.add_batch(
                        attention_map, patches_batch[torchio.LOCATION]
                    )
                else:
                    _, _, output, _ = result

                if params["problem_type"] == "segmentation":
                    aggregator.add_batch(
                        output.detach().cpu(), patches_batch[torchio.LOCATION]
                    )
                else:
                    if torch.is_tensor(output):
                        # this probably needs customization for classification (majority voting or median, perhaps?)
                        output_prediction += output.detach().cpu()
                    else:
                        output_prediction += output

            # save outputs
            if params["problem_type"] == "segmentation":
                output_prediction = aggregator.get_output_tensor()
                output_prediction = output_prediction.unsqueeze(0)
                if params["save_output"]:
                    img_for_metadata = torchio.ScalarImage(
                        tensor=subject["1"]["data"].squeeze(0),
                        affine=subject["1"]["affine"].squeeze(0),
                    ).as_sitk()
                    pred_mask = output_prediction.numpy()
                    # perform postprocessing before reverse one-hot encoding here
                    for postprocessor in params["data_postprocessing"]:
                        for _class in range(0, params["model"]["num_classes"]):
                            pred_mask[0, _class, ...] = global_postprocessing_dict[
                                postprocessor
                            ](pred_mask[0, _class, ...], params)
                    # '0' because validation/testing dataloader always has batch size of '1'
                    pred_mask = reverse_one_hot(
                        pred_mask[0], params["model"]["class_list"]
                    )
                    pred_mask = np.swapaxes(pred_mask, 0, 2)

                    # perform postprocessing after reverse one-hot encoding here
                    for postprocessor in params[
                        "data_postprocessing_after_reverse_one_hot_encoding"
                    ]:
                        pred_mask = global_postprocessing_dict[postprocessor](
                            pred_mask, params
                        )

                    # if jpg detected, convert to 8-bit arrays
                    ext = get_filename_extension_sanitized(subject["1"]["path"][0])
                    if ext in [
                        ".jpg",
                        ".jpeg",
                        ".png",
                    ]:
                        pred_mask = pred_mask.astype(np.uint8)

                    ## special case for 2D
                    if image.shape[-1] > 1:
                        result_image = sitk.GetImageFromArray(pred_mask)
                    else:
                        result_image = sitk.GetImageFromArray(pred_mask.squeeze(0))
                    result_image.CopyInformation(img_for_metadata)

                    # this handles cases that need resampling/resizing
                    if "resample" in params["data_preprocessing"]:
                        result_image = resample_image(
                            result_image,
                            img_for_metadata.GetSpacing(),
                            interpolator=sitk.sitkNearestNeighbor,
                        )
                    # Create the subject directory if it doesn't exist in the
                    # current_output_dir directory
                    os.makedirs(
                        os.path.join(current_output_dir, "testing"),
                        exist_ok=True,
                    )
                    os.makedirs(
                        os.path.join(
                            current_output_dir, "testing", subject["subject_id"][0]
                        ),
                        exist_ok=True,
                    )

                    path_to_save = os.path.join(
                        current_output_dir,
                        "testing",
                        subject["subject_id"][0],
                        subject["subject_id"][0] + "_seg" + ext,
                    )

                    sitk.WriteImage(
                        result_image,
                        path_to_save,
                    )
            else:
                # final regression output
                output_prediction = output_prediction / len(patch_loader)
                if calculate_overall_metrics:
                    predictions_array[batch_idx] = (
                        torch.argmax(output_prediction[0], 0).cpu().item()
                    )
                if params["save_output"]:
                    outputToWrite += (
                        str(epoch)
                        + ","
                        + subject["subject_id"][0]
                        + ","
                        + str(output_prediction)
                        + "\n"
                    )

            # get the final attention map and save it
            if params["medcam_enabled"] and params["model"]["type"] == "torch":
                attention_map = attention_map_aggregator.get_output_tensor()
                for i, n in enumerate(attention_map):
                    model.save_attention_map(
                        n.squeeze(), raw_input=image[i].squeeze(-1)
                    )

            output_prediction = output_prediction.squeeze(-1)
            if is_inference and is_classification:
                logits_list.append(output_prediction)
                subject_id_list.append(subject.get("subject_id")[0])

            # we cast to float32 because float16 was causing nan
            if label_ground_truth is not None:
                # this is for RGB label
                if label_ground_truth.shape[0] == 3:
                    label_ground_truth = label_ground_truth[0, ...].unsqueeze(0)
                # we always want the ground truth to be in the same format as the prediction
                label_ground_truth = label_ground_truth.unsqueeze(0)
                if label_ground_truth.shape[-1] == 1:
                    label_ground_truth = label_ground_truth.squeeze(-1)
                final_loss, final_metric = get_loss_and_metrics(
                    image,
                    label_ground_truth,
                    output_prediction.to(torch.float32),
                    params,
                )
                if params["verbose"]:
                    print(
                        "Full image " + mode + ":: Loss: ",
                        final_loss,
                        "; Metric: ",
                        final_metric,
                        flush=True,
                    )

                # # Non network validing related
                # loss.cpu().data.item()
                total_epoch_valid_loss += final_loss.cpu().item()
                for metric in final_metric.keys():
                    if isinstance(total_epoch_valid_metric[metric], list):
                        if len(total_epoch_valid_metric[metric]) == 0:
                            total_epoch_valid_metric[metric] = np.array(
                                final_metric[metric]
                            )
                        else:
                            total_epoch_valid_metric[metric] += np.array(
                                final_metric[metric]
                            )
                    else:
                        total_epoch_valid_metric[metric] += final_metric[metric]

        if label_ground_truth is not None:
            if params["verbose"]:
                # For printing information at halftime during an epoch
                if ((batch_idx + 1) % (len(valid_dataloader) / 2) == 0) and (
                    (batch_idx + 1) < len(valid_dataloader)
                ):
                    print(
                        "\nHalf-Epoch Average " + mode + " loss : ",
                        total_epoch_valid_loss / (batch_idx + 1),
                    )
                    for metric in params["metrics"]:
                        if isinstance(total_epoch_valid_metric[metric], np.ndarray):
                            to_print = (
                                total_epoch_valid_metric[metric] / (batch_idx + 1)
                            ).tolist()
                        else:
                            to_print = total_epoch_valid_metric[metric] / (
                                batch_idx + 1
                            )
                        print(
                            "Half-Epoch Average " + mode + " " + metric + " : ",
                            to_print,
                        )

    if params["medcam_enabled"] and params["model"]["type"] == "torch":
        model.disable_medcam()
        params["medcam_enabled"] = False

    if label_ground_truth is not None:
        average_epoch_valid_loss = total_epoch_valid_loss / len(valid_dataloader)
        print("     Epoch Final   " + mode + " loss : ", average_epoch_valid_loss)
        # get overall stats for classification
        if calculate_overall_metrics:
            average_epoch_valid_metric = overall_stats(
                predictions_array, ground_truth_array, params
            )
        average_epoch_valid_metric = print_and_format_metrics(
            average_epoch_valid_metric,
            total_epoch_valid_metric,
            params["metrics"],
            mode,
            len(valid_dataloader),
        )

    else:
        average_epoch_valid_loss, average_epoch_valid_metric = 0, {}

    if scheduler is not None:
        if params["scheduler"]["type"] in [
            "reduce_on_plateau",
            "reduce-on-plateau",
            "plateau",
            "reduceonplateau",
        ]:
            scheduler.step(average_epoch_valid_loss)
        else:
            scheduler.step()

    # write the predictions, if appropriate
    if params["save_output"]:
        if is_inference and is_classification and logits_list:
            class_list = [str(c) for c in params["model"]["class_list"]]
            logit_tensor = torch.cat(logits_list)
            current_fold_dir = params["current_fold_dir"]
            logit_tensor = logit_tensor.detach().cpu().numpy()
            columns = ["SubjectID"] + class_list
            logits_df = pd.DataFrame(columns=columns)
            logits_df.SubjectID = subject_id_list
            logits_df[class_list] = logit_tensor

            logits_file = os.path.join(current_fold_dir, "logits.csv")
            if os.path.isfile(logits_file):
                logits_file = os.path.join(
                    current_fold_dir, "logits_" + get_unique_timestamp() + ".csv"
                )
            logits_df.to_csv(logits_file, index=False, sep=",")

        if "value_keys" in params:
            file = open(file_to_write, "w")
            file.write(outputToWrite)
            file.close()

    return average_epoch_valid_loss, average_epoch_valid_metric



import torch
from torchmetrics import (
    F1Score,
    Precision,
    Recall,
    JaccardIndex,
    Accuracy,
    Specificity,
)
from GANDLF.utils.tensor import one_hot
from GANDLF.utils.generic import (
    determine_classification_task_type,
    define_average_type_key,
    define_multidim_average_type_key,
)


def generic_function_output_with_check(predicted_classes, label, metric_function):
    if torch.min(predicted_classes) < 0:
        print(
            "WARNING: Negative values detected in prediction, cannot compute torchmetrics calculations."
        )
        return torch.zeros((1), device=predicted_classes.device)
    else:
        # I need to do this with try-except, otherwise for binary problems it will
        # raise and error as the binary metrics do not have .num_classes
        # attribute.
        # https://github.com/Lightning-AI/torchmetrics/blob/v1.1.2/src/torchmetrics/classification/accuracy.py#L31-L146 link to example from BinaryAccuracy.
        try:
            max_clamp_val = metric_function.num_classes - 1
        except AttributeError:
            max_clamp_val = 1
        predicted_new = torch.clamp(predicted_classes.cpu().int(), max=max_clamp_val)
        predicted_new = predicted_new.reshape(label.shape)
        return metric_function(predicted_new, label.cpu().int())


def generic_torchmetrics_score(output, label, metric_class, metric_key, params):
    task = determine_classification_task_type(params)
    num_classes = params["model"]["num_classes"]
    predicted_classes = output
    if params["problem_type"] == "classification":
        predicted_classes = torch.argmax(output, 1)
    elif params["problem_type"] == "segmentation":
        label = one_hot(label, params["model"]["class_list"])
    metric_function = metric_class(
        task=task,
        num_classes=num_classes,
        threshold=params["metrics"][metric_key]["threshold"],
        average=define_average_type_key(params, metric_key),
        multidim_average=define_multidim_average_type_key(params, metric_key),
    )

    return generic_function_output_with_check(
        predicted_classes.cpu().int(), label.cpu().int(), metric_function
    )


def recall_score(output, label, params):
    return generic_torchmetrics_score(output, label, Recall, "recall", params)


def precision_score(output, label, params):
    return generic_torchmetrics_score(output, label, Precision, "precision", params)


def f1_score(output, label, params):
    return generic_torchmetrics_score(output, label, F1Score, "f1", params)


def accuracy(output, label, params):
    return generic_torchmetrics_score(output, label, Accuracy, "accuracy", params)


def specificity_score(output, label, params):
    return generic_torchmetrics_score(output, label, Specificity, "specificity", params)


def iou_score(output, label, params):
    num_classes = params["model"]["num_classes"]
    predicted_classes = output
    if params["problem_type"] == "classification":
        predicted_classes = torch.argmax(output, 1)
    elif params["problem_type"] == "segmentation":
        label = one_hot(label, params["model"]["class_list"])
    task = determine_classification_task_type(params)
    recall = JaccardIndex(
        task=task,
        num_classes=num_classes,
        average=define_average_type_key(params, "iou"),
        threshold=params["metrics"]["iou"]["threshold"],
    )

    return generic_function_output_with_check(
        predicted_classes.cpu().int(), label.cpu().int(), recall
    )



import sys
import torch


# Dice scores and dice losses
def dice(predicted: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    """
    This function computes a dice score between two tensors.

    Args:
        predicted (torch.Tensor): Predicted value by the network.
        target (torch.Tensor): Required target label to match the predicted with

    Returns:
        torch.Tensor: The computed dice score.
    """
    predicted_flat = predicted.flatten()
    label_flat = target.flatten()
    intersection = (predicted_flat * label_flat).sum()

    dice_score = (2.0 * intersection + sys.float_info.min) / (
        predicted_flat.sum() + label_flat.sum() + sys.float_info.min
    )

    return dice_score


def mcc(predictions: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    """
    This function computes the Matthews Correlation Coefficient (MCC) between two tensors. Adapted from https://github.com/kakumarabhishek/MCC-Loss/blob/main/loss.py.

    Args:
        predictions (torch.Tensor): The predicted value by the network.
        targets (torch.Tensor): Required target label to match the predicted with

    Returns:
        torch.Tensor: The computed MCC score.
    """
    tp = torch.sum(torch.mul(predictions, targets))
    tn = torch.sum(torch.mul((1 - predictions), (1 - targets)))
    fp = torch.sum(torch.mul(predictions, (1 - targets)))
    fn = torch.sum(torch.mul((1 - predictions), targets))

    numerator = torch.mul(tp, tn) - torch.mul(fp, fn)
    # Adding epsilon to the denominator to avoid divide-by-zero errors.
    denominator = (
        torch.sqrt(
            torch.add(tp, 1, fp)
            * torch.add(tp, 1, fn)
            * torch.add(tn, 1, fp)
            * torch.add(tn, 1, fn)
        )
        + torch.finfo(torch.float32).eps
    )

    return torch.div(numerator.sum(), denominator.sum())


def generic_loss_calculator(
    predicted: torch.Tensor,
    target: torch.Tensor,
    num_class: int,
    loss_criteria,
    weights: list = None,
    ignore_class: int = None,
    loss_type: int = 0,
) -> torch.Tensor:
    """
    This function computes the mean class dice score between two tensors

    Args:
        predicted (torch.Tensor): Predicted generally by the network
        target (torch.Tensor): Required target label to match the predicted with
        num_class (int): Number of classes (including the background class)
        loss_criteria (function): Loss function to use
        weights (list, optional): Dice weights for each class (excluding the background class), defaults to None
        ignore_class (int, optional): Class to ignore, defaults to None
        loss_type (int, optional): Type of loss to compute, defaults to 0. The options are:
            0: no loss, normal dice calculation
            1: dice loss, (1-dice)
            2: log dice, -log(dice)

    Returns:
        torch.Tensor: Mean Class Dice score
    """
    accumulated_loss = 0
    # default to a ridiculous value so that it is ignored by default
    ignore_class = -1e10 if ignore_class is None else ignore_class

    for class_index in range(num_class):
        if class_index != ignore_class:
            current_loss = loss_criteria(
                predicted[:, class_index, ...], target[:, class_index, ...]
            )

            # subtract from 1 because this is supposed to be a loss
            default_loss = 1 - current_loss
            if loss_type == 2 or loss_type == "log":
                # negative because we want positive losses, and add epsilon to avoid infinities
                current_loss = -torch.log(current_loss + torch.finfo(torch.float32).eps)
            else:
                current_loss = default_loss

            # multiply by appropriate weight if provided
            if weights is not None:
                current_loss = current_loss * weights[class_index]

            accumulated_loss += current_loss

    if weights is None:
        accumulated_loss /= num_class

    return accumulated_loss


def MCD_loss(
    predicted: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function computes the Dice loss between two tensors. These weights should be the penalty weights, not dice weights.

    Args:
        predicted (torch.Tensor): The predicted value by the network.
        target (torch.Tensor): Required target label to match the predicted with
        params (dict): Dictionary of parameters

    Returns:
        torch.Tensor: The computed MCC loss.
    """
    return generic_loss_calculator(
        predicted,
        target,
        len(params["model"]["class_list"]),
        dice,
        params["penalty_weights"],
        None,
        1,
    )


def MCD_log_loss(
    predicted: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function computes the Dice loss between two tensors with log. These weights should be the penalty weights, not dice weights.

    Args:
        predicted (torch.Tensor): The predicted value by the network.
        target (torch.Tensor): Required target label to match the predicted with
        params (dict): Dictionary of parameters

    Returns:
        torch.Tensor: The computed MCC loss.
    """
    return generic_loss_calculator(
        predicted,
        target,
        len(params["model"]["class_list"]),
        dice,
        params["penalty_weights"],
        None,
        2,
    )


def MCC_loss(
    predicted: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function computes the Matthews Correlation Coefficient (MCC) loss between two tensors. These weights should be the penalty weights, not dice weights.

    Args:
        predicted (torch.Tensor): The predicted value by the network.
        target (torch.Tensor): Required target label to match the predicted with
        params (dict): Dictionary of parameters

    Returns:
        torch.Tensor: The computed MCC loss.
    """
    return generic_loss_calculator(
        predicted,
        target,
        len(params["model"]["class_list"]),
        mcc,
        params["penalty_weights"],
        None,
        1,
    )


def MCC_log_loss(
    predicted: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function computes the Matthews Correlation Coefficient (MCC) loss between two tensors with log. These weights should be the penalty weights, not dice weights.

    Args:
        predicted (torch.Tensor): The predicted value by the network.
        target (torch.Tensor): Required target label to match the predicted with
        params (dict): Dictionary of parameters

    Returns:
        torch.Tensor: The computed MCC loss.
    """
    return generic_loss_calculator(
        predicted,
        target,
        len(params["model"]["class_list"]),
        mcc,
        params["penalty_weights"],
        None,
        2,
    )


def tversky_loss(
    predicted: torch.Tensor, target: torch.Tensor, alpha: float = 0.5, beta: float = 0.5
) -> torch.Tensor:
    """
    This function calculates the Tversky loss between two tensors.

    Args:
        predicted (torch.Tensor): Predicted generally by the network
        target (torch.Tensor): Required target label to match the predicted with
        alpha (float, optional): Weight of false positives. Defaults to 0.5.
        beta (float, optional): Weight of false negatives. Defaults to 0.5.

    Returns:
        torch.Tensor: Computed Tversky Loss
    """
    # Move this part later to parameter parsing, no need to check every time
    assert 0 <= alpha <= 1, f"Invalid alpha value: {alpha}"
    assert 0 <= beta <= 1, f"Invalid beta value: {beta}"
    assert 0 <= alpha + beta <= 1, f"Invalid alpha and beta values: {alpha}, {beta}"

    predicted_flat = predicted.contiguous().view(-1)
    target_flat = target.contiguous().view(-1)

    true_positives = (predicted_flat * target_flat).sum()
    false_positives = ((1 - target_flat) * predicted_flat).sum()
    false_negatives = (target_flat * (1 - predicted_flat)).sum()

    numerator = true_positives
    denominator = true_positives + alpha * false_positives + beta * false_negatives
    score = (numerator + sys.float_info.min) / (denominator + sys.float_info.min)

    loss = 1 - score
    return loss


def MCT_loss(
    predicted: torch.Tensor, target: torch.Tensor, params: dict = None
) -> torch.Tensor:
    """
    This function calculates the Multi-Class Tversky loss between two tensors.

    Args:
        predicted (torch.Tensor): Predicted generally by the network
        target (torch.Tensor): Required target label to match the predicted with
        params (dict, optional): Additional parameters for computing loss function, including weights for each class

    Returns:
        torch.Tensor: Computed Multi-Class Tversky Loss
    """

    acc_tv_loss = 0
    num_classes = predicted.shape[1]

    for i in range(num_classes):
        curr_loss = tversky_loss(predicted[:, i, ...], target[:, i, ...])
        if params is not None and params.get("penalty_weights") is not None:
            curr_loss = curr_loss * params["penalty_weights"][i]
        acc_tv_loss += curr_loss

    if params is not None and params.get("penalty_weights") is None:
        acc_tv_loss /= num_classes

    return acc_tv_loss


def KullbackLeiblerDivergence(mu, logvar, params=None):
    """
    Calculates the Kullback-Leibler divergence between two Gaussian distributions.

    Args:
        mu (torch.Tensor): The mean of the first Gaussian distribution
        logvar (torch.Tensor): The logarithm of the variance of the first Gaussian distribution
        params (dict, optional): A dictionary of optional parameters

    Returns:
        torch.Tensor: The computed Kullback-Leibler divergence
    """
    loss = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp(), dim=-1)
    return loss.mean()


def FocalLoss(
    predicted: torch.Tensor, target: torch.Tensor, params: dict = None
) -> torch.Tensor:
    """
    This function calculates the Focal loss between two tensors.

    Args:
        predicted (torch.Tensor): Predicted generally by the network
        target (torch.Tensor): Required target label to match the predicted with
        params (dict, optional): Additional parameters for computing loss function, including weights for each class

    Returns:
        torch.Tensor: Computed Focal Loss
    """
    gamma = 2.0
    size_average = True
    if isinstance(params["loss_function"], dict):
        gamma = params["loss_function"].get("gamma", 2.0)
        size_average = params["loss_function"].get("size_average", True)

    def _focal_loss(preds, target, gamma, size_average=True):
        """
        Internal helper function to calculate focal loss for a single class.

        Args:
            preds (torch.Tensor): predicted generally by the network
            target (torch.Tensor): Required target label to match the predicted with
            gamma (float): The gamma value for focal loss
            size_average (bool, optional): Whether to average the loss across the batch. Defaults to True.

        Returns:
            torch.Tensor: Computed focal loss for a single class.
        """
        ce_loss = torch.nn.CrossEntropyLoss(reduce=False)
        logpt = ce_loss(preds, target)
        pt = torch.exp(-logpt)
        loss = ((1 - pt) ** gamma) * logpt
        return_loss = loss.sum()
        if size_average:
            return_loss = loss.mean()
        return return_loss

    acc_focal_loss = 0
    num_classes = predicted.shape[1]

    for i in range(num_classes):
        curr_loss = _focal_loss(
            predicted[:, i, ...], target[:, i, ...], gamma, size_average
        )
        if params is not None and params.get("penalty_weights") is not None:
            curr_loss = curr_loss * params["penalty_weights"][i]
        acc_focal_loss += curr_loss

    return acc_focal_loss



import os
import shutil
import yaml
import docker
import tarfile
import io
import sysconfig

# import copy

deploy_targets = [
    "docker",
    #'singularity',
    #'onefile'
]

mlcube_types = ["model", "metrics"]


def run_deployment(
    mlcubedir,
    outputdir,
    target,
    mlcube_type,
    entrypoint_script=None,
    configfile=None,
    modeldir=None,
    requires_gpu=None,
):
    """
    Run the deployment of the model.

    Args:
        mlcubedir (str): The path to the mlcube directory.
        outputdir (str): The path to the output directory.
        target (str): The target to deploy to.
        mlcube_type (str): Either 'model' or 'metrics'
        entrypoint_script (str): The path of entrypoint script. Only used for metrics and inference
        configfile (str, Optional): The path to the configuration file. Required for models
        modeldir (str, Optional): The path to the model directory. Required for models
        requires_gpu (str, Optional): Whether the model requires GPU. Required for models

    Returns:
        bool: True if the deployment was successful, False otherwise.
    """
    assert (
        target in deploy_targets
    ), f"The deployment target {target} is not a valid target."

    assert (
        mlcube_type in mlcube_types
    ), f"The mlcube type {mlcube_type} is not a valid type."

    if not os.path.exists(outputdir):
        os.makedirs(outputdir, exist_ok=True)

    assert not os.path.isfile(
        outputdir
    ), f"Output location {outputdir} exists but is a file, not a directory."

    if mlcube_type == "model":
        assert os.path.exists(
            modeldir
        ), f"Error: The model path {modeldir} does not exist."

        assert os.path.isdir(
            modeldir
        ), f"The model path {modeldir} exists but is not a directory."

        assert os.path.exists(
            configfile
        ), f"The config file {configfile} does not exist."

    if entrypoint_script:
        assert os.path.exists(
            entrypoint_script
        ), f"Error: The script path {entrypoint_script} does not exist."

    if target.lower() == "docker":
        result = deploy_docker_mlcube(
            mlcubedir, outputdir, entrypoint_script, configfile, modeldir, requires_gpu
        )
        assert result, "Something went wrong during platform-specific deployment."

    return True


def deploy_docker_mlcube(
    mlcubedir,
    outputdir,
    entrypoint_script=None,
    config=None,
    modeldir=None,
    requires_gpu=None,
):
    """
    Deploy the docker mlcube of the model or metrics calculator.

    Args:
        mlcubedir (str): The path to the mlcube directory.
        outputdir (str): The path to the output directory.
        entrypoint_script (str): The path of entrypoint script. Only used for metrics and inference
        config (str, Optional): The path to the configuration file. Required for models
        modeldir (str, Optional): The path to the model directory. Required for models
        requires_gpu (str, Optional): Whether the model requires GPU. Required for models
    """
    mlcube_config_file = os.path.join(mlcubedir, "mlcube.yaml")
    assert os.path.exists(mlcubedir) and os.path.exists(
        mlcube_config_file
    ), "MLCube Directory: This does not appear to be a valid MLCube directory."

    output_workspace_folder = os.path.join(outputdir, "workspace")
    mlcube_workspace_folder = os.path.join(mlcubedir, "workspace")

    os.makedirs(output_workspace_folder, exist_ok=True)
    if os.path.exists(mlcube_workspace_folder):
        shutil.copytree(
            mlcube_workspace_folder,
            output_workspace_folder,
            dirs_exist_ok=True,
        )

    if config is not None:
        shutil.copyfile(config, os.path.join(output_workspace_folder, "config.yml"))

    # First grab the existing the mlcube config
    if modeldir is not None:
        # we can use that as an indicator if we are doing model or metrics deployment
        mlcube_config = get_model_mlcube_config(
            mlcube_config_file, requires_gpu, entrypoint_script
        )
    else:
        mlcube_config = get_metrics_mlcube_config(mlcube_config_file, entrypoint_script)

    output_mlcube_config_path = os.path.join(outputdir, "mlcube.yaml")
    with open(output_mlcube_config_path, "w") as f:
        f.write(yaml.dump(mlcube_config, sort_keys=False, default_flow_style=False))

    # This tag will be modified later by our deployment mechanism
    docker_image = mlcube_config["docker"]["image"]

    # Run the mlcube_docker configuration process, forcing build from local repo
    gandlf_root = os.path.realpath(os.path.dirname(__file__) + "/../../")
    site_packages_dir = sysconfig.get_path("purelib")
    symlink_location = ""
    if (
        gandlf_root == site_packages_dir
    ):  # Installed via pip, not as editable source install, extra work is needed
        setup_files = ["setup.py", ".dockerignore", "pyproject.toml", "MANIFEST.in"]
        dockerfiles = [
            item
            for item in os.listdir(gandlf_root)
            if os.path.isfile(os.path.join(gandlf_root, item))
            and item.startswith("Dockerfile-")
        ]
        entrypoints = [
            item
            for item in os.listdir(gandlf_root)
            if os.path.isfile(os.path.join(gandlf_root, item))
            and item.startswith("gandlf_")
        ]
        for file in setup_files + dockerfiles + entrypoints:
            shutil.copy(
                os.path.join(gandlf_root, file),
                os.path.join(gandlf_root, "GANDLF", file),
            )
        if not os.path.exists(os.path.join(gandlf_root, "GANDLF", "GANDLF")):
            # point to same package directory, acts as a recursive location for the GaNDLF package
            symlink_location = os.path.join(gandlf_root, "GANDLF", "GANDLF")
            os.symlink("./", os.path.join(gandlf_root, "GANDLF", "GANDLF"))
        gandlf_root = os.path.join(gandlf_root, "GANDLF")

    print(os.listdir(gandlf_root))

    # Requires mlcube_docker python package to be installed with scripts available
    # command_to_run = "mlcube_docker configure --platform=docker  -Pdocker.build_strategy=always" + " --mlcube=" + os.path.realpath(mlcubedir) + " -Pdocker.build_context=" + gandlf_root
    command_to_run = (
        "mlcube_docker configure --platform=docker  -Prunner.build_strategy=always"
        + " --mlcube="
        + os.path.realpath(mlcubedir)
        + " -Prunner.build_context="
        + gandlf_root
    )

    print("Running MLCube configuration with the following command:")
    print(command_to_run)
    print(
        "If this is your first GaNDLF deployment, this may take longer than usual while image layers are built."
    )

    assert (
        os.system(command_to_run) == 0
    ), "mlcube_docker configuration failed. Check output for more details."

    # Container is made at this point, the recursive symlink no longer needs to exist
    if os.path.exists(symlink_location) and os.path.islink(symlink_location):
        os.unlink(symlink_location)

    # If mlcube_docker configuration worked, the image is now present in Docker so we can manipulate it.
    docker_client = docker.from_env()
    print("Connected to the docker service.")
    container = docker_client.containers.create(docker_image)

    # Embed mlcube config
    embed_asset(output_mlcube_config_path, container, "embedded_mlcube.yml")

    # Embed modeldir if available
    print("Attempting to embed the model...")
    embed_asset(modeldir, container, "embedded_model")

    # Embed config if available
    embed_asset(config, container, "embedded_config.yml")

    # Embed entrypoint script if available
    embed_asset(entrypoint_script, container, "entrypoint.py")

    # Commit the container to the same tag.
    docker_repo = docker_image.split(":")[0]
    docker_tag = docker_image.split(":")[1]
    container.commit(repository=docker_repo, tag=docker_tag)

    print(
        f"The updated container was committed successfully. It is available under Docker as: {docker_image} ."
    )
    print(
        f"This image should be distributed with the MLCube directory created at {outputdir}."
    )
    print(
        "You may now push this image (e.g. to Docker Hub) as normal. By doing so, the user will need only the MLCube directory in order to pull and run the image."
    )
    print(
        f"To run this container as an MLCube, you (or the end-user) should invoke the MLCube runner with --mlcube={outputdir} ."
    )
    print(
        "Otherwise, it will function as a standard docker image of GaNDLF, but with possibly embedded model/config files."
    )
    print("Deployment finished successfully!")
    return True


def get_metrics_mlcube_config(mlcube_config_file, entrypoint_script):
    """
    This function is used to get the metrics from mlcube config file.

    Args:
        mlcube_config_file (str): The path of mlcube config file.
        entrypoint_script (str): The path of entrypoint script.
    """
    mlcube_config = None
    with open(mlcube_config_file, "r") as f:
        mlcube_config = yaml.safe_load(f)
    if entrypoint_script:
        # modify the entrypoint to run a custom script
        mlcube_config["tasks"]["evaluate"]["entrypoint"] = "python3.9 /entrypoint.py"
    mlcube_config["docker"]["build_strategy"] = "auto"
    return mlcube_config


def get_model_mlcube_config(mlcube_config_file, requires_gpu, entrypoint_script):
    """
    This function returns the mlcube config for the model.

    Args:
        mlcube_config_file (str): Path to mlcube config file.
        requires_gpu (bool): Whether the model requires GPU.
        entrypoint_script (str): The path of entrypoint script. Only used for infer task
    """
    mlcube_config = None
    with open(mlcube_config_file, "r") as f:
        mlcube_config = yaml.safe_load(f)
    mlcube_config["docker"]["build_strategy"] = "auto"

    # modify for the embedded-model image.
    old_train_output_modeldir = mlcube_config["tasks"]["train"]["parameters"][
        "outputs"
    ].pop("modeldir", None)
    mlcube_config["tasks"]["infer"]["parameters"]["inputs"].pop("modeldir", None)
    mlcube_config["tasks"]["train"]["parameters"]["inputs"].pop("config", None)
    # Currently disabled because we've decided exposing config-on-inference complicates the MLCube use case.
    # mlcube_config["tasks"]["infer"]["parameters"]["inputs"].pop("config", None)

    # Change output so that each task always places secondary output in the workspace
    mlcube_config["tasks"]["train"]["parameters"]["outputs"][
        "output_path"
    ] = old_train_output_modeldir

    # Change entrypoints to point specifically to the embedded model and config
    mlcube_config["tasks"]["train"]["entrypoint"] = (
        mlcube_config["tasks"]["train"]["entrypoint"]
        + " --modeldir /embedded_model/"
        + " --config /embedded_config.yml"
    )
    mlcube_config["tasks"]["infer"]["entrypoint"] = (
        mlcube_config["tasks"]["infer"]["entrypoint"]
        + " --modeldir /embedded_model/"
        + " --config /embedded_config.yml"
    )

    # Change some configuration if GPU is required by default
    if requires_gpu:
        mlcube_config["platform"]["accelerator_count"] = 1
        mlcube_config["tasks"]["train"]["entrypoint"] = mlcube_config["tasks"]["train"][
            "entrypoint"
        ] = mlcube_config["tasks"]["train"]["entrypoint"].replace(
            "--device cpu", "--device cuda"
        )
        mlcube_config["tasks"]["infer"]["entrypoint"] = mlcube_config["tasks"]["infer"][
            "entrypoint"
        ] = mlcube_config["tasks"]["infer"]["entrypoint"].replace(
            "--device cpu", "--device cuda"
        )

    if entrypoint_script:
        # modify the infer entrypoint to run a custom script.
        device = "cuda" if requires_gpu else "cpu"
        mlcube_config["tasks"]["infer"][
            "entrypoint"
        ] = f"python3.9 /entrypoint.py --device {device}"

    return mlcube_config
    # Duplicate training task into one from reset (must be explicit) and one that resumes with new data
    # In either case, the embedded model will not change persistently.
    # The output in workspace will be the result of resuming training with new data on the embedded model.
    # This is currently disabled -- "reset" and "resume" seem to behave strangely in these conditions.
    # mlcube_config["tasks"]["training_from_reset"] = copy.deepcopy(
    #    mlcube_config["tasks"]["training"]
    # )
    # mlcube_config["tasks"]["training_from_reset"]["entrypoint"] = (
    #    mlcube_config["tasks"]["training_from_reset"]["entrypoint"] + " --reset True"
    # )
    # mlcube_config["tasks"]["training"]["entrypoint"] = (
    #    mlcube_config["tasks"]["training"]["entrypoint"] + " --resume True"
    # )


def embed_asset(asset, container, asset_name):
    """
    This function embeds an asset into a container.

    Args:
        asset (str): The path to the asset to embed.
        container (docker.models.containers.Container): The container to embed the asset into.
        asset_name (str): The name of the asset to embed.
    """
    # Tarball the modeldir, config.yml and mlcube.yaml into memory, insert into container
    if asset is not None:
        if os.path.exists(asset):
            asset_file_io = io.BytesIO()
            with tarfile.open(fileobj=asset_file_io, mode="w|gz") as tar:
                tar.add(os.path.realpath(asset), arcname=asset_name)
            asset_file_io.seek(0)
            container.put_archive("/", asset_file_io)
            asset_file_io.close()


## TODO!
## If implemented, even users who can't install singularity/docker/python could run frozen models.
## https://nuitka.net/ (Nuitka) can generate one-file binaries from python scripts, and can include additional files in the bundled self-extracting filesystem.
## https://pyinstaller.org/en/stable/ (PyInstaller) can do the same, albeit with a different approach.
## By bundling both GaNDLF and the required model/config files in one executable, we could have true drag-and-drop deployment of GaNDLF models.
## This is ideal, because then not even docker / singularity engines are needed.
## However, both of the above have occasional problems with certain packages (such as SimpleITK or Torch).
## In both cases, deployments need to be done on the corresponding platform/architecture, including separate CUDA/Torch versions.
## To implement this will need a lot of experimentation with each compiler/freezer and our own packages.
# def deploy_onefile(modeldir, config, outputdir):
#    raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Implementation of Adaptive gradient clipping
"""

import torch


def unitwise_norm(x, norm_type=2.0):
    """
    Computes the norm of a tensor x, where the norm is applied across all dimensions except the first one.

    Args:
        x (torch.Tensor): Input tensor.
        norm_type (float): The type of norm to compute (default: 2.0).

    Returns:
        torch.Tensor: The norm of the tensor.
    """
    if x.ndim <= 1:
        return x.norm(norm_type)
    else:
        # This should work for nn.ConvNd (where N is the number of dimensions)
        # and nn.Linear where number of output dimensions is first
        # in the weight/kernel tensor passed.
        # This might need something for a few unplanned scenarios
        # but we dont have to deal with them here.
        return x.norm(norm_type, dim=tuple(range(1, x.ndim)), keepdim=True)


def adaptive_gradient_clip_(parameters, clip_factor=0.01, eps=1e-3, norm_type=2.0):
    """
    Performs adaptive gradient clipping on the parameters of a PyTorch model.

    Args:
        parameters (list of torch.Tensor): The parameters to be clipped.
        clip_factor (float): The factor by which to clip the gradients (default: 0.01).
        eps (float): A small value added to the norm to avoid division by zero (default: 1e-3).
        norm_type (float): The type of norm to compute (default: 2.0).

    Adaptive Gradient Clipping
    Original implementation of Adaptive Gradient Clipping derived from
    An impl of AGC, as per (https://arxiv.org/abs/2102.06171):


    Paper Name : High-Performance Large-Scale Image Recognition Without Normalization
    Authors : Andrew Brock and Soham De and Samuel L. Smith and Karen Simonyan
    Published : arXiv preprint arXiv: 2021

    Code references:
      * Official JAX impl (paper authors): https://github.com/deepmind/deepmind-research/tree/master/nfnets
      * Phil Wang's PyTorch gist: https://gist.github.com/lucidrains/0d6560077edac419ab5d3aa29e674d5c
    """
    # If parameter is not a list, make it one so that you can iterate over it
    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]

    # Parse through the list of parameters
    for p in parameters:
        if p.grad is None:
            continue
        p_data = p.detach()
        g_data = p.grad.detach()
        max_norm = (
            unitwise_norm(p_data, norm_type=norm_type).clamp_(min=eps).mul_(clip_factor)
        )
        grad_norm = unitwise_norm(g_data, norm_type=norm_type)
        clipped_grad = g_data * (max_norm / grad_norm.clamp(min=1e-6))
        new_grads = torch.where(grad_norm < max_norm, g_data, clipped_grad)
        p.grad.detach().copy_(new_grads)



from GANDLF.utils.modelbase import get_modelbase_final_layer
from GANDLF.metrics import surface_distance_ids


def populate_header_in_parameters(parameters, headers):
    """
    This function populates the parameters with information from the header in a common manner

    Args:
        parameters (dict): The parameters passed by the user yaml.
        headers (dict): The CSV headers dictionary.

    Returns:
        dict: Combined parameter dictionary containing header information
    """
    # initialize common parameters based on headers
    parameters["headers"] = headers
    # ensure the number of output classes for model prediction is working correctly

    if len(headers["predictionHeaders"]) > 0:
        parameters["model"]["num_classes"] = len(headers["predictionHeaders"])

    # initialize model type for processing: if not defined, default to torch
    parameters["model"]["type"] = parameters["model"].get("type", "torch")

    if parameters["model"]["type"] == "openvino" and parameters["model"][
        "architecture"
    ] in ["brain_age", "sdnet"]:
        print(
            "Only PyTorch for inference is supported for the current network topology: {0}.".format(
                parameters["model"]["architecture"]
            )
        )
        parameters["model"]["type"] = "torch"

    # initialize number of channels for processing
    parameters["model"]["num_channels"] = parameters["model"].get(
        "num_channels", len(headers["channelHeaders"])
    )

    parameters["problem_type"] = find_problem_type(
        parameters, get_modelbase_final_layer(parameters["model"]["final_layer"])
    )

    # if the problem type is classification/segmentation, ensure the number of classes are picked from the configuration
    if parameters["problem_type"] != "regression":
        parameters["model"]["num_classes"] = len(parameters["model"]["class_list"])

    return parameters


def find_problem_type(parameters, model_final_layer):
    """
    This function determines the type of problem at hand - regression, classification or segmentation

    Args:
        parameters (dict): The parameters passed by the user yaml.
        model_final_layer (model_final_layer): The final layer of the model. If None, the model is for regression.

    Returns:
        str: The problem type (regression/classification/segmentation).
    """
    # check if regression/classification has been requested
    classification_phrases = [
        "classification_but_not_softmax",
        "logits",
        "classification_without_softmax",
        "classification_with_sigmoid",
    ]
    headersFromCSV = parameters["headers"]
    class_list_exist = "class_list" in parameters["model"]
    if (
        class_list_exist
        and parameters["model"]["final_layer"].lower() in classification_phrases
    ):
        return "classification"

    if len(headersFromCSV["predictionHeaders"]) > 0:
        if model_final_layer is None:
            return "regression"
        else:
            return "classification"
    else:
        return "segmentation"


def find_problem_type_from_parameters(parameters):
    """
    This function determines the type of problem at hand - regression, classification or segmentation

    Args:
        parameters (dict): The parameters passed by the user yaml.

    Returns:
        str: The problem type (regression/classification/segmentation).
    """
    # check if classification has been requested
    final_layer_classification_phrases = [
        "classification_but_not_softmax",
        "logits",
        "classification_without_softmax",
        "classification_with_sigmoid",
    ]
    model_final_layer = parameters["model"]["final_layer"]
    if isinstance(model_final_layer, str):
        model_final_layer = model_final_layer.lower()
        # check if regression has been requested
        if model_final_layer in ["regression", "none"]:
            return "regression"
        elif model_final_layer in final_layer_classification_phrases:
            return "classification"
        else:
            for metric in parameters["metrics"]:
                if isinstance(metric, str):
                    metric = metric.lower()
                    if (
                        metric
                        in ["dice", "dice_score", "dc", "dice_coefficient"]
                        + surface_distance_ids
                    ):
                        return "segmentation"
            return "classification"
    elif model_final_layer is None:
        return "regression"


def populate_channel_keys_in_params(data_loader, parameters):
    """
    Function to read channel key information from specified data loader

    Args:
        data_loader (torch.DataLoader): The data loader to query key information from.
        parameters (dict): The parameters passed by the user yaml.

    Returns:
        dict: Updated parameters that include key information
    """
    # using train_loader makes this slower as train loader contains full augmentations
    batch = next(iter(data_loader))
    all_keys = list(batch.keys())
    channel_keys = []
    value_keys = []
    label_keys = []
    for item in all_keys:
        if item.isnumeric():
            channel_keys.append(item)
        elif "value" in item:
            value_keys.append(item)
        elif "label" in item:
            label_keys.append(item)
    parameters["channel_keys"] = channel_keys
    if value_keys:
        parameters["value_keys"] = value_keys
    if label_keys:
        parameters["label_keys"] = label_keys

    return parameters
